import hashlib
import socket
import time
import logging
import rconnet

logger = logging.getLogger(rconnet.__name__)

class TCPClientException(Exception):
    pass

class TCPClientStatuses:
    DISABLED = 0
    CONNECTED = 1
    RECONNECTING = 2

class TCPClient(object):

    def __init__(self, address: str,
                 password: int,
                 port: int=4711,
                 on_connect = None,
                 on_disconnect = None,
                 on_close = None,
                 on_status = None,
                 ):

        self.socket = None
        self.address = address
        self.password = password
        self.port = port
        self.on_connect = on_connect,
        self.on_disconnect = on_disconnect,
        self.on_close = on_close,
        self.on_status = on_status,
        self._status = TCPClientStatuses.DISABLED
        self._inr = 0

    def __debug(self, text):
        logger.debug(f"[{self.address} {self.port}] {text}")

    def __info(self, text):
        logger.info(f"[{self.address} {self.port}] {text}")

    def __error(self, text):
        logger.error(f"[{self.address} {self.port}] {text}")

    def _exec_event(self, name, *args):
        method = getattr(self, name)
        if type(method) == tuple: method = method[0]
        if method is not None:
            self.__debug(f"calling an event `{name}`")
            method(*args)

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, status):
        self.__info(f"A new status has been set {status}")
        self._status = status
        self._exec_event("on_status", status)

    def start(self):
        self.__info("Opens a socket connection ...")
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect((self.address, self.port))
        self.__info("Connected")
        self._pre_init()
        self.status = TCPClientStatuses.CONNECTED
        self._exec_event("on_connect")
        return True

    def _pre_init(self):
        self.__info("Performing pre-initialization ...")
        self.pre_init()
        self.__info("Successful pre-initialization")

    def pre_init(self):
        welcResponse = ""
        while 1:
            data = self.socket.recv(1024)
            welcResponse += data.decode("utf-8")
            eosPos = welcResponse.find('\n\n')
            if eosPos != -1: break;

        prefix = '### Digest seed: '
        seedPos = welcResponse.find(prefix)
        seed = ""

        if seedPos == -1: raise Exception("Authorization is not possible")

        seedPos += len(prefix)
        seedPosEnd = welcResponse.find('\n', seedPos)
        seed = welcResponse[seedPos:seedPosEnd].encode("utf-8")

        md5Obj = hashlib.md5()
        md5Obj.update(seed)
        md5Obj.update(self.password.encode("utf-8"))
        passHash = md5Obj.hexdigest()

        self.socket.send(('\x02' + 'login ' + passHash + '\n').encode("utf-8"))
        data = self.socket.recv(1024).decode("utf-8")
        if "Authentication successful, rcon ready." not in data: raise Exception(data)

    def is_closed(self):
        return self.status == TCPClientStatuses.DISABLED

    def close(self):
        if not self.socket: return None
        self.__info("Closing the connection ...")
        self.status = TCPClientStatuses.DISABLED
        self.socket.close()
        self._exec_event("on_close")

    def send(self, data):
        self.__debug(f"send `{data}`")
        self.socket.send(data)

    def recv(self):
        result = self.socket.recv(2048)
        self.__debug(f"recv `{result}`")
        return result

    def rcon_invoke(self, command):
        if not self.socket: raise Exception("The client is not connected")
        self.send(('\x02' + command + '\n').encode("utf-8"))
        self._inr += 1
        result = ""
        done = False
        while not done:
            try:
                data = self.recv()
                if data is None: raise Exception("Client has terminated the current connection. ")
            except Exception as error:
                self.status = TCPClientStatuses.DISABLED
                self._exec_event("on_disconnect", error)
                raise Exception(error)
            for c in data:
                if c == 0x4:
                    done = True
                    break
                result += chr(c)
        if len(result) > 0 and result[-1] == "\n": result = result[:-1].strip()
        if result == "": return None
        return result

class TCPListener(TCPClient):

    def __init__(self, *args, commands:list = [], on_recv:object = None, **kwargs):
        super().__init__(*args, **kwargs)
        self._commands = commands
        self.on_recv = on_recv

    def pre_init(self):
        super().pre_init()
        for command in self._commands:
            self.socket.send(('\x02' + command + '\n').encode("utf-8"))

    def start(self):
        while True:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                self.socket.connect((self.address, self.port))
                self._pre_init()
                self.status = TCPClientStatuses.CONNECTED
                self._exec_event("on_connect")
            except Exception as error:
                time.sleep(3)
                print(error)
                continue

            while 1:
                try:
                    data = self.socket.recv(2048)
                    if data is None: raise Exception("Client has terminated the current connection. ")
                except Exception as error:
                    self._exec_event("on_disconnect", error)
                    break
                result = data
                done = False
                if result[-1] == 0x4: done = True
                while not done:
                    data = self.socket.recv(2048)
                    for c in data:
                        if c == 0x4:
                            done = True
                            break
                        result += chr(c)
                self._exec_event("on_recv", result)



class Server(object):

	def __init__(self, si):

		self.si = []

		for i in si.split('\t'):
			try:
				value = int(i)
				self.si.append(value)
				continue
			except:
				pass

			try:
				value = float(i)
				self.si.append(value)
				continue
			except:
				pass

			try:
				value = str(i)
				self.si.append(value)
			except:
				pass

		self.reservedSlots = self.si[29]
		self.wallTime = self.si[28]
		self.team2 = self.si[27]
		self.team1 = self.si[26]
		self.ranked = self.si[25]
		self.autoBalanceTeam = self.si[24]
		self.timeLimit = self.si[23]
		self.worldSize = self.si[22]
		self.modDir = self.si[21]
		self.gameMode = self.si[20]
		self.timeLeft = self.si[19]
		self.roundTime = self.si[18]
		self.team2_null = self.si[17]
		self.team2_tickets = self.si[16]
		self.team2_startTickets = self.si[15]
		self.team2_TicketState = self.si[14]
		self.team2_Name = self.si[13]
		self.team1_null = self.si[12]
		self.team1_tickets = self.si[11]
		self.team1_startTickets = self.si[10]
		self.team1_TicketState = self.si[9]
		self.team1_Name = self.si[8]
		self.serverName = self.si[7]
		self.nextMapName = self.si[6]
		self.currentMapName = self.si[5]
		self.joiningPlayers = self.si[4]
		self.connectedPlayers = self.si[3]
		self.maxPlayers = self.si[2]
		self.currentGameStatus = self.si[1]
		self.version = self.si[0]

class Player():

	def __init__(self, player):
		self.player = player

		self.index = self.player[0]
		self.getName = self.player[1]
		self.getTeam = self.player[2]
		self.getPing = self.player[3]
		self.isConnected = self.player[4]
		self.isValid = self.player[5]
		self.isRemote = self.player[6]
		self.isAIPlayer = self.player[7]
		self.isAlive = self.player[8]
		self.isManDown = self.player[9]
		self.getProfileId = self.player[10]
		self.isFlagHolder = self.player[11]
		self.getSuicide = self.player[12]
		self.getTimeToSpawn = self.player[13]
		self.getSquadId = self.player[14]
		self.isSquadLeader = self.player[15]
		self.isCommander = self.player[16]
		self.getSpawnGroup = self.player[17]
		self.getAddress = self.player[18]
		self.scoreDamageAssists = self.player[19]
		self.scorePassengerAssists = self.player[20]
		self.scoreTargetAssists = self.player[21]
		self.scoreRevives = self.player[22]
		self.scoreTeamDamages = self.player[23]
		self.scoreTeamVehicleDamages = self.player[24]
		self.scoreCpCaptures = self.player[25]
		self.scoreCpDefends = self.player[26]
		self.scoreCpAssists = self.player[27]
		self.scoreCpNeutralizes = self.player[28]
		self.scoreCpNeutralizeAssists = self.player[29]
		self.scoreSuicides = self.player[30]
		self.scoreKills = self.player[31]
		self.scoreTKs = self.player[32]
		self.vehicleType = self.player[33]
		self.kitTemplateName = self.player[34]
		self.kiConnectedAt = self.player[35]
		self.deaths = self.player[36]
		self.score = self.player[37]
		self.vehicleName = self.player[38]
		self.rank = self.player[39]
		self.position = self.player[40]
		self.idleTime = self.player[41]
		self.keyhash = self.player[42]
		self.punished = self.player[43]
		self.timesPunished = self.player[44]
		self.timesForgiven = self.player[45]

class Players():

	def __init__(self, pl):
		self.pl = {}
		self.pi = []

		self.players = None
		self.playersTeam1 = 0
		self.playersTeam2 = 0

		self.scoreTeam1 = 0
		self.scoreTeam2 = 0

		self.killsTeam1 = 0
		self.killsTeam2 = 0

		self.deathsTeam1 = 0
		self.deathsTeam2 = 0

		self.pingTeam1 = 0.00
		self.pingTeam2 = 0.00

		self.kdrTeam1 = 0.00
		self.kdrTeam2 = 0.00

		self.pingTeam1_all = 0
		self.pingTeam2_all = 0

		for pi_ in pl.split('\r'):
			self.pi = []
			for p in pi_.split('\t'):
				try:
					value = int(p)
					self.pi.append(value)
					continue
				except:
					pass

				try:
					value = float(p)
					self.pi.append(value)
					continue
				except:
					pass

				try:
					value = str(p)
					self.pi.append(value)
				except:
					pass
			if self.pi[0] == '': break
			self.pl[self.pi[0]] = self.pi

		self.players = self.pl

		for i, y in self.players.items():
			try:
				if y[2] == 1:
					self.playersTeam1 += 1
					self.scoreTeam1 += y[37]
					self.killsTeam1 += y[31]
					self.deathsTeam1 += y[36]
					self.pingTeam1_all += y[3]
			except:
				pass

			try:
				if y[2] == 2:
					self.playersTeam2 += 1
					self.scoreTeam2 += y[37]
					self.killsTeam2 += y[31]
					self.deathsTeam2 += y[36]
					self.pingTeam2_all += y[3]
			except:
				pass

		try:
			self.kdrTeam1 = float('{:.2f}'.format(self.killsTeam1 / self.deathsTeam1))
			self.kdrTeam2 = float('{:.2f}'.format(self.killsTeam2 / self.deathsTeam2))

			self.pingTeam1 = float('{:.2f}'.format(self.pingTeam1_all / self.playersTeam1))
			self.pingTeam2 = float('{:.2f}'.format(self.pingTeam2_all / self.playersTeam2))
		except ZeroDivisionError as e:
			pass

class Bf2cc():
	def __init__(self, client):
		self.__client = client

	def pause(self):
		self.__client._has_full_drive()
		return self.__client.rcon_invoke("bf2cc pause")

	def unpause(self):
		self.__client._has_full_drive()
		return self.__client.rcon_invoke("bf2cc unpause")

	@property
	def info(self):
		self.__client._has_full_drive()
		return self.__client.rcon_invoke("bf2cc check")

	@property
	def server(self):
		self.__client._has_full_drive()
		data = self.__client.rcon_invoke("bf2cc si")
		return Server(data)

	#@property
	#def players(self):
	#	self.__client._has_full_drive()
	#	data = self.__client.rcon_invoke("bf2cc pl")
	#	return Server(data)


from .default import Default
from .modules.modmanager.bf2cc import Bf2cc
import re

class ModManager(Default):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.bf2cc = Bf2cc(self)

    def start(self):
        super().start()
        if self.rcon_invoke("bf2cc check") == "rcon: unknown command: 'bf2cc'":
            raise Exception("Modmanager is not installed on the server")

    def list_modules(self):
        modules = {}
        pattern = r"^(.*?)\s+v(\d+\.\d+)\s+\(\s+(.*?)\s+\)$"
        lines = self.rcon_invoke("mm listModules")
        for line in lines.split("\n"):
            match = re.match(pattern, line.strip())
            if match:
                modules[match.group(1)] = {
                    "version": match.group(2),
                    "status": match.group(3)
                }

        return modules

    def config(self):
        pattern = r'(?P<section>\w+)\.(?P<option>\w+)\s+(?:(?P<value_int>\d+)|\"(?P<value_str>[^\"]+)\")'
        sections = {}

        data = self.rcon_invoke("mm printRunningConfig")
        for line in data.split("\n"):
            line = line.strip()

            if line.startswith("#"):
                continue

            match = re.match(pattern, line)

            if not match:
                continue

            section = match.group('section')
            option = match.group('option')
            value = match.group('value_int') or match.group('value_str')

            if value.isdigit(): value = int(value)

            sections[section] = sections.get(section, {})

            if option.startswith("add") or option == "loadModule":
                sections[section][option] = sections[section].get(option, [])
                sections[section][option].append(value)
            else:
                sections[section][option] = sections[section].get(option, value)

        return sections

    def reload_module(self, module:str):
        return self.rcon_invoke("mm reloadModule %s" % module)

    def start_module(self, module:str):
        return self.rcon_invoke("mm startModule %s" % module)

    def shutdown_module(self, module:str):
        return self.rcon_invoke("mm shutdownModule %s" % module)

    def load_module(self, module:str):
        return self.rcon_invoke("mm loadModule %s" % module)

    def save_config(self):
        return self.rcon_invoke("mm saveConfig")

    def set_param(self, module:str, option:str, value:str):
        return self.rcon_invoke("mm setParam %s %s %s" % (module, option, value))